"""
##########################################
BlasterTK_FV module functions and members.

Developped for the Blaster plugin by:
	Daniel Barbeau Vasquez, copyright 2005
	Released under the GPL.
	V 1.00
	Inspired by Mariano's "Hotspots"
	script.
##########################################
"""

####             IMPORT OF BLASTER MODULES AND FRIENDS    ####
import Blender, sys
from Blender import Draw, BGL, Window, Image, Text, Scene, Object
####      END OF IMPORT OF BLASTER MODULES AND FRIENDS    ####


__all__ = ["handle_event", "init2", "modif_detect", \
           "get_area_info", "draw_icons", "icon", "main_draw", "main_event"]


#Global declarations:
Blender.G = {}
Blender.G["LOI"] = []
Blender.G["SObj"] = {}
Blender.G["debug"] = False
Blender.G["Hbut"] = {}
Blender.G["Icons"] = {}
#The following imports can only be done here: avoids circular imports
import user_functions
from Blender import G
#End

def init2():
	"""This function takes care of script links.
	Scene links are used to do save/load of script settings
	Is called by BlasterDraw.py on init."""
	#print "doing init2"
	
	####Creating path to scripts.
	script = Blender.Get("scriptsdir") + "/blaster"

	####Is BlasterTK_LI.py already loaded in blender?
	txt_list = []
	for txt in Text.Get():
		txt_list.append(txt.getName())
	
	if "BlasterTK_LI.py" not in txt_list:
		text = Text.Load(script + "/BlasterTK_LI.py")
		if  Scene.GetCurrent().getScriptLinks("OnSave") == None:
			Scene.GetCurrent().addScriptLink(text.name, "OnSave")
			Scene.GetCurrent().addScriptLink(text.name, "OnLoad")
		elif "BlasterTK_LI.py" not in Scene.GetCurrent().getScriptLinks("OnSave"):
			Scene.GetCurrent().addScriptLink(text.name, "OnSave")
			Scene.GetCurrent().addScriptLink(text.name, "OnLoad")
	else:
		text = Text.Get("BlasterTK_LI.py")
		if  Scene.GetCurrent().getScriptLinks("OnSave") == None:
			Scene.GetCurrent().addScriptLink(text.name, "OnSave")
			Scene.GetCurrent().addScriptLink(text.name, "OnLoad")
		elif "BlasterTK_LI.py" not in Scene.GetCurrent().getScriptLinks("OnSave"):
			Scene.GetCurrent().addScriptLink(text.name, "OnSave")
			Scene.GetCurrent().addScriptLink(text.name, "OnLoad")		
		
	return

###################### DRAW SPACE FUNCTION ######################
def main_draw():
	"""Main entry point for 3DView redraw"""
	
	
	#print "main_draw"
	# LOADING Scene Link File
	init2()
	
	#Draw Functions also need to be aware of mesh addition/deletion	
	modif_detect()
	
	#OPENGL INIT
	BGL.glDepthFunc(BGL.GL_LESS)			# The Type Of Depth Test To Do   (I hate these GL stuffs)
	BGL.glLoadIdentity()					# Reset The Projection Matrix
	BGL.glMatrixMode(BGL.GL_MODELVIEW)
	BGL.glEnable(BGL.GL_BLEND)
	BGL.glBlendFunc(BGL.GL_SRC_ALPHA, BGL.GL_ONE_MINUS_SRC_ALPHA)
	
	#retreiving screen data
	scr = get_area_info()
	
	if len(Object.GetSelected()) <> 0:
		CurrAct = Object.GetSelected()[0]
	else: pass
	
	if len(Object.GetSelected()) == 1:# and CurrAct.getType() == "Mesh":
		dict = zipit(G["Icons"], G["SObj"][CurrAct.name]["icons"])
		draw_icons(dict) #can only be called once : thus zipit()
	else:
		draw_icons(G["Icons"])
	return
		
###################### EVENT SPACE FUNCTION ######################
def main_event():
	"""Main entry point for event capturing"""

	#print "main_event"
	evt = Blender.event
	if len(Object.GetSelected()) <> 0:
		CurrAct = Object.GetSelected()[0]
		
	if len(Object.GetSelected()) == 1:# and CurrAct.getType() == "Mesh":
		dict = zipit(G["Icons"], G["SObj"][CurrAct.name]["icons"])
		handle_event(evt, dict) #can only be called once : thus zipit()
	else:
		handle_event(evt, G["Icons"])

	return

def zipit(dict, previous = None):
	"""function to merge two dictionnaries together while preserving order
	arguments:
		- dict : dictionnary to append
		- previous : (optional) dictionnary to which dict will be appended"""
	appended = {}
	if previous == None:
		appended = {}
	else:
		appended = previous
		
	for i in dict.iteritems():
		appended[i[0]] = i[1]
	return appended

###################### LOW LEVEL FUNCTIONS ######################
def extend_buttons(Area, button):
	"""Function to convert the icon's list of 
	parameters into one	more understandable by Python.
	Parameters are:
			-- Area : as returned by get_area_info(). Needed to correctly place icon and hotspot
					{"height" : int, "width" : int, "id": int, "AreaVerts" : [int, int, int, int]}
			-- button : data conerning button to create 
				[str Type, str Name, int Corner, tuple(MarginX, MarginY), \
				bool Object-Relative, bool Has_Right_Click, bool Conserve_Hidden_Data]
				
	Returns: Expanded icon Dictionnary
	Called by icon.__init__()
			  icon.autorefresh()"""

	#print "extend_buttons"
	lwidth, lheight = Area["width"], Area["height"]
	Vardict = {}
	
	#sorry for this thing but I'm lazy tonight:
	but = button
	
	type = but[0]
	#image loading
	if type == "unique":
		path = Blender.Get("scriptsdir") + "/blaster/pics/" + but[1] + ".png"
		downpath = Blender.Get("scriptsdir") + "/blaster/pics/" + "un" + but[1] + ".png"
		image_1 = Image.Load(path)
		image_2 = Image.Load(downpath)
	elif type == "binary":
		path = Blender.Get("scriptsdir") + "/blaster/pics/" + but[1] + ".png"
		unpath = Blender.Get("scriptsdir") + "/blaster/pics/" + "un" + but[1] + ".png"
		image_1 = Image.Load(path)
		image_2 = Image.Load(unpath)
	elif type == "ternary":
		path = Blender.Get("scriptsdir") + "/blaster/pics/" + but[1] + ".png"
		path2 = Blender.Get("scriptsdir") + "/blaster/pics/" + "bis_" + but[1] + ".png"
		unpath = Blender.Get("scriptsdir") + "/blaster/pics/" + "un" + but[1] + ".png"
		image_1 = Image.Load(path)
		image_2 = Image.Load(path2)
		image_3 = Image.Load(unpath)
	
	#Positionning the BLHcorner of the icon in global screen area		
	if but[2] == 1: # Bottom Left Hand corner of area
		positionx = but[3][0]
		positiony = but[3][1]
	elif but[2] == 2: # Bottom Right Hand corner of area
		positionx = lwidth - image_1.size[0] - but[3][0]
		positiony = but[3][1]
	elif but[2] == 3: # Top Right Hand corner of area
		positionx = lwidth - image_1.size[0] - but[3][0]
		positiony = lheight - image_1.size[1] - but[3][1]
	elif but[2] == 4: # Top Left Hand corner of area
		positionx = but[3][0]
		positiony = lheight - image_1.size[1] - but[3][1]
	
	#function identifier creator for "exec"
	if type == "unique":
		do = but[1] + "()"
	elif type == "binary":
		do = but[1] + "()"
		undo = "un" + but[1] + "()"
	elif type == "ternary":
		do = but[1] + "()"
		do2 = "bis_" + but[1] + "()"
		undo = "un" + but[1] + "()"
		
	#action for right click  ... right mouse button doesn't induce icon modification... simple case.
	if but[5] == True:
		right = True , "rclk" + but[1] + "()"
	else:
		right = False, None
	
	####FINAL STEP#### TODO : maybe move Hotspots to their own dictionnary: keep track of corresponding button
	if type == "unique": #hackish case... the icon will change and revert when click is released.
		Vardict["Type"] = type
		Vardict["Image1"] = image_1, do
		Vardict["Image2"] = image_2, 0
		Vardict["Right"] = right
		Vardict["Orig"] = (positionx,positiony)
		Vardict["HotrangeX"] = (positionx + Area["AreaVerts"][0], positionx + image_2.size[0] + Area["AreaVerts"][0])
		Vardict["HotrangeY"] = (positiony + Area["AreaVerts"][1], positiony + image_1.size[1] + Area["AreaVerts"][1])
		Vardict["Name"] = but[1]
	elif type == "binary": #icons of different states must be the same size >> easier if this behaviour is to change, do it here
		Vardict["Type"] = type
		Vardict["Image1"] = image_1, do
		Vardict["Image2"] = image_2, undo
		Vardict["Right"] = right
		Vardict["Orig"] = (positionx,positiony)
		Vardict["HotrangeX"] = (positionx + Area["AreaVerts"][0], positionx + image_2.size[0] + Area["AreaVerts"][0])
		Vardict["HotrangeY"] = (positiony + Area["AreaVerts"][1], positiony + image_1.size[1] + Area["AreaVerts"][1])
		Vardict["Name"] = but[1]
	elif type == "ternary":
		Vardict["Type"] = type
		Vardict["Image1"] = image_1, do
		Vardict["Image2"] = image_2, do2
		Vardict["Image3"] = image_3, undo
		Vardict["Right"] = right
		Vardict["Orig"] = (positionx,positiony)
		Vardict["HotrangeX"] = (positionx + Area["AreaVerts"][0], positionx + image_2.size[0] + Area["AreaVerts"][0])
		Vardict["HotrangeY"] = (positiony + Area["AreaVerts"][1], positiony + image_1.size[1] + Area["AreaVerts"][1])
		Vardict["Name"] = but[1]
	return Vardict

def handle_event(evt, dict):
	"""Distributes the events through BlasterTK's classes and
	functions or sends back event to Blender.
		Parameters are:
			-- evt : Blender's captured event
			-- dict : the dict that should be considered for the event.

		Returns: normally, 3DView Redraw through Queue
		Called by BlasterEvent, whenever needed

		Note: will change to handle internally the different mouse events
			and events on non-object-relative icons"""

	#print "handle_event on %s\n" % str(dict),

	for ico in dict:
		ico = dict[ico]
		current = ico.autoRefresh()
		if ico.mouse_on(current) == True:
			ico.do_action(evt, current)
			loopbacktoblender(evt, True)
			return
		else:
			continue
	loopbacktoblender(evt, False)
	return

def loopbacktoblender(evt, result):
	if result == False:
		Blender.event = evt
		Window.QRedrawAll()
		return
	else:
		Blender.event = None
		Window.QRedrawAll()
		return

def modif_detect():
	"""Detects addition or deletion of mesh objects in the scene
	and updates global Blender.G["SObj"] dictionnary
		No parameters
		No returned object (might change to return local value)
	
		Called by BlasterDraw, BlasterTK_LI and BlasterEvent"""

	#print "modif_detect"
	#deletion detection: updates Blaster's own list.
	tempodict = {}
	for mystobj in Object.Get():
		if G["SObj"].has_key(mystobj.name):
			tempodict[mystobj.name] = G["SObj"][mystobj.name]
	G["SObj"] = tempodict
	#filtering: adding new meshes in the structure.	
	for scobj in Object.Get():
		if G["SObj"].has_key(scobj.name) == 0:
#		if scobj.getType() == "Mesh" and \
#		G["SObj"].has_key(scobj.name) == 0:
			G["SObj"][scobj.name] = {}
			G["SObj"][scobj.name]["icons"] = {} 
		else:
			pass
	#appending icons to objects in Blaster's G structure or loads them in non-object-relative structure..
	for i in G["LOI"]:
		if i[4] == True:
			for obj in G["SObj"].iterkeys():
				if len(G["SObj"][obj]["icons"]) < len(G["LOI"]):
					G["SObj"][obj]["icons"][i[1]] = icon(i[0],i[1],i[2],i[3],i[4],i[5],i[6])
				else:
					continue
		elif G["Icons"].has_key(i[1]) == False:
			G["Icons"][i[1]] = icon(i[0],i[1],i[2],i[3],i[4],i[5],i[6])
	return

def get_area_info():
	"""Collects info concerning the current 3DView.
	Shouldn't be forgotten that this is executed serially throughout
	all 3DViews having the BlasterTK space handler loaded.
	This means that the data going through extend_buttons() is overwritten
	as many times per screen redraw as the number 3D views
	
	Called by icon.__init__()
			  icon.autorefresh()
			  icon.draw()
			  icon.mouse_on()
			  icon.do_action()"""

	#print "get_area_info"
	Area = dict()
	Area["width"], Area["height"] = Window.GetAreaSize()
	Area["id"] = Window.GetAreaID()
	areas = Window.GetScreenInfo()
	for area in areas:
		if area["id"] == Area["id"]:
			Area["AreaVerts"] = area["vertices"]
	return Area


def draw_icons(dict):
	"""Propagates redraw event through all icons declared in BlasterTK.
	Parameters are: 
		-- dict : which dict to draw
	Returns: Screen updates
	"""
	#print "draw_icons"		
	for i in dict.itervalues():
		current = i.autoRefresh()
		i.draw(current)
	return

#BEGIN class icon
class icon:
	"""My precious :)
		an icon is an instance of this class"""
		
	def __init__(self, p_type, p_name, p_corner, p_margin, p_obdep, p_hasrght, p_cons):
		"""Initialise all attributes needed by class. No read-only attribute as it makes
		programming a bit harder, but this may change so nobody messes with fragile internals"""
		#print "%s._init_\n" % p_name,
		self.Type = p_type
		self.Name = p_name
		self.Corner = p_corner
		self.Margin = p_margin
		self.Obdep = p_obdep
		self.Hasrght = p_hasrght
		self.Cons = p_cons
		self.state = 0
		self.scre = get_area_info()										
		self.paramlist = [self.Type, self.Name, self.Corner, self.Margin, self.Obdep, self.Hasrght, self.Cons, self.state]
		self.extend_version = extend_buttons(self.scre, self.paramlist)
		return

	def getState(self):
		"""Method for reteving icon.state (used by OnSave scriptlink in BlasterTK_LI)
			Returns :int icon.state"""
		#print "%s.getState\n" % self.Name,
		return self.state

	def setState(self,state):
		"""Method for changing icon.state (used by OnLoad scriptlink in BlasterTK_LI)"""
		#print "%s.setState\n" % self.Name,
		self.state = state
		return
	
	def autoRefresh(self):
		"""Isn"t really that automated. It simple refreshes icon.extend_version
		as returned by extend_buttons"""
		#print "%s.autoRefresh\n" % self.Name,
		return extend_buttons(get_area_info(), self.paramlist)

	
	def autoChangeState(self, type):
		"""automatically changes icon.state on mouse click according to icon type"""
		#print "%s.autoChangeState\n" % self.Name,
		if type == "unique":
			pass
		elif type == "binary":
			if self.state == 0:
				self.state = 1
				return
			elif self.state == 1:
				self.state = 0
				return
		elif type == "ternary":
			if self.state == 0:
				self.state = 1
				return
			elif self.state == 1:
				self.state = 2
				return
			elif self.state == 2:
				self.state = 0
				return
		return
	
	def draw(self, current):
		"""automated method that draws the picture according to
		icon state and type"""
		#print "%s.draw\n" % self.Name,
		#Rules for handling unique icons. Careful, hacks inside.
		if current["Type"] == "unique":
			#if off, turn it on
			if self.state == 1:
				Draw.Image(current["Image2"][0], \
				current["Orig"][0], \
				current["Orig"][1])
			#if on, turn it off
			else:
				Draw.Image(current["Image1"][0], \
				current["Orig"][0], \
				current["Orig"][1])
				return		
		#Rules for handling binary icons.
		elif current["Type"] == "binary":
			#if off, turn it on
			if self.state == 0:
				Draw.Image(current["Image1"][0], \
				current["Orig"][0], \
				current["Orig"][1])
				return
				#if on, turn it off
			else:
				Draw.Image(current["Image2"][0], \
				current["Orig"][0], \
				current["Orig"][1])
				return
		#Rules for handling ternary icons.		
		elif current["Type"] == "ternary":
			#if off, turn it on
			if self.state == 0:
				Draw.Image(current["Image1"][0], \
				current["Orig"][0], \
				current["Orig"][1])
				return
			#if on, turn it on2
			elif self.state == 1:
				Draw.Image(current["Image2"][0], \
				current["Orig"][0], \
				current["Orig"][1])
				return
			#if on, turn it off
			else:
				Draw.Image(current["Image3"][0], \
				current["Orig"][0], \
				current["Orig"][1])
				return
		return
	
		
	def mouse_on(self, current):
		"""detects if event happened over the icon.
			Returns: bool True if test is positive."""
		#print "%s.mouse_on\n" % self.Name,
		mouse_coords = Window.GetMouseCoords()
		if (current["HotrangeX"][0]< mouse_coords[0] < current["HotrangeX"][1]) \
			and (current["HotrangeY"][0]< mouse_coords[1] < current["HotrangeY"][1]):
			return True
		else:
			return False
		
	def do_action(self, evt, current):
		"""automated method to call the icon's associated functions and change the
		icon's state if successful."""
		#print "%s.do_action\n" % self.Name,
		if evt == Draw.LEFTMOUSE: #leftmouse click
			if current["Type"] == "unique":
				event = Window.GetMouseButtons()
				test = self.mouse_on(current)
				while evt <> 0 and test == True :
					self.setState(1)
					Window.Redraw()
					evt = Window.GetMouseButtons()
					test = self.mouse_on(current)
				else:
					if test == True:
						exec("user_functions." + current["Image1"][1])
						self.setState(0)
					else:
						self.setState(0)
				return
			#Rules for handling binary icons.		
			elif current["Type"] == "binary":
				if self.state == 0:
					exec("user_functions." + current["Image1"][1])
					self.autoChangeState(self.Type)
					return
				elif self.state == 1:
					exec("user_functions." + current["Image2"][1])
					self.autoChangeState(self.Type)
					return
			#Rules for handling ternary icons.		
			elif current["Type"] == "ternary":
				if self.state == 0:
					exec("user_functions." + current["Image1"][1])
					self.autoChangeState(self.Type)
					return
				elif self.state == 1:
					exec("user_functions." + current["Image2"][1])
					self.autoChangeState(self.Type)
					return
				elif self.state == 2:
					exec("user_functions." + current["Image3"][1])
					self.autoChangeState(self.Type)
					return
		elif evt == Draw.RIGHTMOUSE: #right click is easy :)
			if current["Right"][0] == True:
				exec("user_functions." + current["Right"][1])
				return
			else:
				pass
		return
#END class icon
